unit DupFileEliminator;
 (* Eliminate files from one directory tree when they are respectively
  identical to files in identical positions in another tree *)

  interface
    uses
      Classes,
      Controls (* crHourglass, tCursor *),
      ExtCtrls (* tLabeledEdit *),
      Forms    (* Application, tCloseAction, tForm, Screen *),
      StdCtrls (* tButton, tCheckBox, tGroupBox, tListBox *),
      Windows;

    type
      tfmDupEliminator = class(TForm)
          btnClear:    tButton;
          btnGo:       tButton;
          btnNewToOld: tButton;
          ckbxAddCleanedTag:     tCheckBox;
          ckbxListFileDeletions: tCheckBox;
          grpNewTree: tGroupBox;
            btnBrowseForNewRoot: tButton;
            lbedNewDirectoryTree: tLabeledEdit;
          grpOldTree: tGroupBox;
            btnBrowseForOldRoot: tButton;
            lbedOldDirectoryTree: tLabeledEdit;
          lbxExceptions: tListBox;
          lbxLog:        tListBox;
    lbDeletedBytes: TLabel;

          procedure btnBrowseForNewRootClick(Sender: tObject);
          procedure btnBrowseForOldRootClick(Sender: tObject);
          procedure btnClearClick           (Sender: tObject);
          procedure btnGoClick              (Sender: tObject);
          procedure btnNewToOldClick        (Sender: tObject);
          procedure FormShow                (Sender: tObject);

          procedure FormClose(Sender: TObject; var Action: TCloseAction);
        private
          FullBackupRoot: string;
          procedure BrowseForDirectory(TheEditBox: tLabeledEdit);
        end (* tfmDupEliminator *);

    var
      fmDupEliminator: tfmDupEliminator;

  implementation
    uses
                                                       {$WARN UNIT_PLATFORM OFF}
      FileCtrl (* SelectDirectory *),
                                                       {$WARN UNIT_PLATFORM ON}
      MyConsts (* PERIOD *),
      Registry (* tRegistry *),
      SysUtils (* DeleteFile, DirectoryExists, faAnyFile, faDirectory,
                    FindClose, FindFirst, FindNext, Format, RemoveDir,
                    RenameFile, tSearchRec, sleep *);
      {$R *.dfm}

    const
      CreateIfMissing = true;
      DFEKey              = 'Software\RRF\DuplicateFileEliminator';
      OldDirKey           = 'OldDirectory';
      NewDirKey           = 'NewDirectory';
      FullBackupRootKey   = 'FullBackupRoot';

    procedure tfmDupEliminator.BrowseForDirectory(TheEditBox: tLabeledEdit);
      (* Starting from the current contents of the edit box, find a directory *)

      const
        NoCreatePermission = [];
        NoHelpContext = 0;

      var
        Tentative: string;

      begin (* tfmDupEliminator.BrowseForDirectory *)
        Tentative := TheEditBox.Text;
        if SelectDirectory(Tentative, NoCreatePermission, NoHelpContext) then
          TheEditBox.Text := Tentative
      end   (* tfmDupEliminator.BrowseForDirectory *);

    procedure tfmDupEliminator.btnBrowseForNewRootClick(Sender: tObject);
      (* usual directory-finding dialog *)

      begin (* tfmDupEliminator.btnBrowseForNewRootClick *)
        BrowseForDirectory(lbedNewDirectoryTree)
      end   (* tfmDupEliminator.btnBrowseForNewRootClick *);

    procedure tfmDupEliminator.btnBrowseForOldRootClick(Sender: tObject);
      (* usual directory-finding dialog *)

      begin (* tfmDupEliminator.btnBrowseForOldRootClick *)
        BrowseForDirectory(lbedOldDirectoryTree)
      end   (* tfmDupEliminator.btnBrowseForOldRootClick *);

    procedure tfmDupEliminator.btnClearClick(Sender: tObject);
      (* cut clutter *)

      begin (* tfmDupEliminator.btnClearClick *)
        lbxLog.Items.Clear;
        lbxExceptions.Items.Clear
      end   (* tfmDupEliminator.btnClearClick *);

    procedure tfmDupEliminator.btnGoClick(Sender: tObject);
      (* Run the old directory tree *)

      var
        BytesDeleted: int64;
        NewRoot, OldRoot: string;

      procedure RunOldDirectory(const OldDirName: string);
        (* recursive; look through files & directories in DirName *)

        const
          MakeLogEntry = true;

        var
          OldCursor: tCursor;
          FindValue: integer;
          CountPast: longint;
          OldSearchRec: tSearchRec;
          NewDirName, NewFileName, OldFileName: string;

{$WARNINGS OFF}  // FindData is specific to a platform
        procedure LogDeletion(SR: tSearchRec;
                              WantLogEntry: boolean = false;
                              WhatDeleted: string = nullstring);
          (* show progress to user *)

          const
            Megabyte = 1 shl 20;
            Gigabyte = 1 shl 30;

          begin (* LogDeletion *)
            with SR.FindData do
              inc(BytesDeleted, nFileSizeHigh shl 32 + nFileSizeLow);
            with lbDeletedBytes do
              if BytesDeleted < Megabyte then
                Caption := Format('deleted %.1f KB', [BytesDeleted / 1000.0])
              else if BytesDeleted < Gigabyte then
                Caption := Format('deleted %.2f MB', [BytesDeleted / Megabyte])
              else
                Caption := Format('deleted %.3f GB', [BytesDeleted / Gigabyte]);

            if WantLogEntry then
              with lbxLog do
                ItemIndex := Items.Add('deleted ' + WhatDeleted)
          end   (* LogDeletion *);

        procedure ProcessFile;
          (* If the new file exists and seems to be identical, delete the old
          file *)

          var
            NewSearchRec: tSearchRec;

          begin (* ProcessFile *)
            if FindFirst(NewFileName, faAnyFile, NewSearchRec) = 0 then
              if NewSearchRec.Time = OldSearchRec.Time then
                if NewSearchRec.Size = OldSearchRec.Size then
                  if NewSearchRec.FindData.nFileSizeHigh = OldSearchRec.FindData.nFileSizeHigh then
                    if NewSearchRec.FindData.nFileSizeLow = OldSearchRec.FindData.nFileSizeLow then
                      if not DeleteFile(OldFileName) then
                        with lbxExceptions do
                          ItemIndex := Items.Add(Format('could not delete %s',
                                                        [OldFileName]))
                      else
                        LogDeletion(NewSearchRec, ckbxListFileDeletions.Checked, OldFileName);

            FindClose(NewSearchRec)
          end   (* ProcessFile *);
{$WARNINGS ON}          

        begin (* RunOldDirectory *)
          OldCursor := Screen.Cursor;
          Screen.Cursor := crHourglass;
          try     (* restore cursor *)
            NewDirName := NewRoot + copy(OldDirName, length(OldRoot)+1, maxint);
            CountPast := 0;
            FindValue := FindFirst(Format('%s\*.*', [OldDirName]),
                                   faAnyFile,
                                   OldSearchRec);
            while FindValue = 0 do
              begin (* process target item *)
                inc(CountPast);
                if CountPast mod 25 = 0 then
                  begin (* allow interruption *)
                    sleep(1);
                    Application.ProcessMessages
                  end   (* allow interruption *);
                if OldSearchRec.Name <> PERIOD then
                  if OldSearchRec.Name <> '..' then
                    begin (* nondegenerate *)
                      OldFileName := Format('%s\%s', [OldDirName, OldSearchRec.Name]);
                      NewFileName := Format('%s\%s', [NewDirName, OldSearchRec.Name]);

                      if (OldSearchRec.Attr and faDirectory) = 0 then
                        ProcessFile
                      else if DirectoryExists(NewFileName) then
                        begin (* run subdir, and delete it if it's now empty *)
                          RunOldDirectory(OldFileName);
                          if RemoveDir(OldFileName) then
                            LogDeletion(OldSearchRec, MakeLogEntry, 'empty directory ' + OldFileName)
                        end (* run subdir, and delete it if it's now empty *)
                    end   (* nondegenerate *);
                    
                FindValue := FindNext(OldSearchRec)
              end   (* process target item *);

            FindClose(OldSearchRec)
          finally (* restore cursor *)
            Screen.Cursor := OldCursor
          end     (* restore cursor *)
        end   (* RunOldDirectory *);

      begin (* tfmDupEliminator.btnGoClick *)
        NewRoot := lbedNewDirectoryTree.Text;
        OldRoot := lbedOldDirectoryTree.Text;
        BytesDeleted := 0;

        if NewRoot <> OldRoot then
          if DirectoryExists(NewRoot) then
            if DirectoryExists(OldRoot) then
              begin (* really do it *)
                RunOldDirectory(OldRoot);
                if ckbxAddCleanedTag.Checked then
                  RenameFile(OldRoot, OldRoot + ' (cleaned)')
              end   (* really do it *);

        with lbxLog do
          ItemIndex := Items.Add('done')
      end   (* tfmDupEliminator.btnGoClick *);

    procedure tfmDupEliminator.btnNewToOldClick(Sender: tObject);
      (* usual case *)

      begin (* tfmDupEliminator.btnNewToOldClick *)
        lbedOldDirectoryTree.Text := lbedNewDirectoryTree.Text
      end   (* tfmDupEliminator.btnNewToOldClick *);

    procedure tfmDupEliminator.FormClose(Sender: tObject;
                                         var Action: tCloseAction);
      (* save the edit boxes *)

      begin (* tfmDupEliminator.FormClose *)
        with tRegistry.Create do
          try     (* using TheRegistry *)
            OpenKey(DFEKey, CreateIfMissing);
            WriteString(NewDirKey, lbedNewDirectoryTree.Text);
            WriteString(OldDirKey, lbedOldDirectoryTree.Text)
          finally (* using TheRegistry *)
            Free
          end     (* using TheRegistry *)
      end   (* tfmDupEliminator.FormClose *);

    procedure tfmDupEliminator.FormShow(Sender: tObject);
      (* preload the edit boxes *)

      begin (* tfmDupEliminator.FormShow *)
        lbDeletedBytes.Caption := nullstring;
        
        with tRegistry.Create do
          try     (* using the registry *)
            if OpenKey(DFEKey, not CreateIfMissing) then
              begin (* get misc registered info *)
                lbedNewDirectoryTree.Text := ReadString(NewDirKey);
                lbedOldDirectoryTree.Text := ReadString(OldDirKey);
                FullBackupRoot            := ReadString(FullBackupRootKey)
              end   (* get misc registered info *)
          finally (* using the registry *)
            Free
          end     (* using the registry *);
      end   (* tfmDupEliminator.FormShow *);

end.
